// This code was converted from part of the Active Template Library.
// Copyright (C) Microsoft Corporation
// All rights reserved.

unit RGSRegistration;

interface

uses
  Windows, Classes;

const
  E_NOT_IN_MAP                    = $0202;
  E_UNEXPECTED_EOS                = $0203;
  E_VALUE_SET_FAILED              = $0204;
  E_EXPECTING_EQUAL               = $0206;
  E_BAD_HKEY                      = $020C;
  E_MISSING_OPENKEY_TOKEN         = $020D;
  E_TYPE_NOT_SUPPORTED            = $020F;
  E_COMPOUND_KEY                  = $0211;

  DirSep            = '\';
  RightBracket      = '}';
  LeftBracket       = '{';
  Quote             = '''';
  cEquals            = '=';
  StringVal       = 'S';
  MultiStringVal  = 'M';
  DwordVal        = 'D';
  BinaryVal       = 'B';
  ValToken        = 'Val';
  ForceRemove     = 'ForceRemove';
  NoRemove        = 'NoRemove';
  Delete          = 'Delete';

type
  TRegObject = class;

  TRegParser = class
  strict protected
  	fRegObj : TRegObject;
  	Cur : PWideChar;
  	procedure SkipWhiteSpace;
  	function NextToken(var Token : string) : HResult;
	  function AddValue(rkParent : HKey; ValueName : string; var Token : string) : HResult;
	  function CanForceRemoveKey(Key : string) : boolean;
	  function HasSubKeys(key : HKEY) : boolean;
  	function HasValues(key : HKEY) : boolean;
	  function RegisterSubkeys(var Token : string; hkParent : HKEY; Register : boolean; Recover : boolean = FALSE) : HResult;
  	function IsSpace(ch : char) : boolean;

  	function GenerateError(val : UINT) : HResult;
  	function SkipAssignment(var Token : string) : HResult;

  	function EndOfVar : boolean;
	  class function StrChr(sz : string; ch : Char) : PWideChar;
	  class function HKeyFromString(Token : string) : HKey;
	  class function ChToByte(const ch : Char) : byte;
	  class function VTFromRegType(ValueType : string; var vt : TVarType) : boolean;
  public
  	constructor Create(RegObj : TRegObject);

	  function PreProcessBuffer(Reg : string; var OutReg : string) : HResult;
	  function RegisterBuffer(Buffer : string; Register : boolean) : HResult;
  end;

  IRegistrarBase = interface
    function AddReplacement(Key: string; Value: string): HResult;
    function ClearReplacements(): HResult;
  end;

  IRegistrar = interface(IRegistrarBase)
    function ResourceRegister(FileName: string; ID: cardinal; sType: string): HResult;
    function ResourceRegisterSz(FileName: string; ID: string; sType: string): HResult;
    function ResourceUnregister(FileName: string; ID: cardinal; sType: string): HResult;
    function ResourceUnregisterSz(FileName: string; ID: string; sType: string): HResult;
    function FileRegister(FileName: string): HResult;
    function FileUnregister(FileName: string): HResult;
    function StringRegister(Data: string): HResult;
    function StringUnregister(Data: string): HResult;
  end;

  TRegObject = class(TInterfacedObject,IRegistrarBase, IRegistrar)
  strict protected
  	RepMap : TStrings;
    fPerUser : boolean;
    function CommonFileRegister(FileName: string; Register: boolean): HResult;
    function RegisterFromResource(FileName: string; ID: PWideChar; sType: string; Register: boolean): HResult;
    function RegisterWithString(Data: string; Register: boolean): HResult;
    function GenerateError(val : cardinal): HResult;  {return DISP_E_EXCEPTION;}
  public
  	destructor Destroy; override;
	  constructor Create(PerUser : boolean);

  	// Map based methods
    function AddReplacement(Key: string; Value: string): HResult;
    function ClearReplacements: HResult;
    function StrFromMap(Key: string): string;

  	// Register via a given mechanism
    function ResourceRegister(FileName: string; ID: cardinal; sType: string): HResult;
    function ResourceRegisterSz(FileName: string; ID: string; sType: string): HResult;
    function ResourceUnregister(FileName: string; ID: cardinal; sType: string): HResult;
    function ResourceUnregisterSz(FileName: string; ID: string; sType: string): HResult;
    function FileRegister(FileName: string): HResult;
    function FileUnregister(FileName: string): HResult;
    function StringRegister(Data: string): HResult;
    function StringUnregister(Data: string): HResult;
    property PerUser : boolean read fPerUser;
  end;


function UpdateRegistryFromResource(hInst : HMODULE; Res : PWideChar;
	Register : boolean; MapEntries : TStrings; PerUser : boolean) : HResult;

function UpdateRegistryFromFile(hInst : HMODULE; Filename : string;
	Register : boolean; MapEntries : TStrings; PerUser : boolean) : HResult;

implementation

uses
  SysUtils, ActiveX;

constructor TRegObject.Create(PerUser : boolean);
begin
  fPerUser := PerUser;
  RepMap := TStringlist.Create;
end;

destructor TRegObject.Destroy;
begin
  ClearReplacements;
  RepMap.Free;
  inherited;
end;

function TRegObject.AddReplacement(Key: string; Value: string): HResult;
begin
	if (Key = '') or (Value = '') then Exit(E_INVALIDARG);
	RepMap.Values[Key] := Value;
	Result := S_OK;
end;

function TRegObject.RegisterFromResource(FileName: string; ID: PWideChar; sType: string; Register: boolean): HResult;
var
	hr : HRESULT;
	parser : TRegParser;
  inst : HMODULE;
	hrscReg : HRSRC;
	hReg: HGLOBAL;
	RegA: AnsiString;
	szReg : string;
begin
  parser := TRegParser.Create(self);

	Inst := LoadLibraryEx(PWideChar(FileName), 0, LOAD_LIBRARY_AS_DATAFILE);

	if (Inst = 0) then
		Exit(HResultFromWin32(GetLastError));
  try
  	hrscReg := FindResource(Inst, ID, PWideChar(sType));

  	if (hrscReg = 0) then
	  begin
		  hr := HResultFromWin32(GetLastError);
  		Exit(hr);
	  end;

  	hReg := LoadResource(Inst, hrscReg);
	  if (hReg = 0) then
  	begin
	  	hr := HResultFromWin32(GetLastError);
  		Exit(hr);
	  end;

  	RegA := PAnsiChar(hReg);
    szReg := string(RegA);

  	hr := parser.RegisterBuffer(szReg, Register);
  finally
    FreeLibrary(Inst);
  end;
	Result := hr;
end;

function TRegObject.ResourceRegister(FileName: string; ID: cardinal; sType: string): HResult;
begin
	Result := RegisterFromResource(FileName, MAKEINTRESOURCE(ID), sType, TRUE);
end;

function TRegObject.ResourceRegisterSz(FileName: string; ID: string; sType: string): HResult;
begin
	if (ID = '') or (sType = '') then Exit(E_INVALIDARG);
	Result := RegisterFromResource(FileName, PWideChar(ID), sType, TRUE);
end;

function TRegObject.ResourceUnregister(FileName: string; ID: cardinal; sType: string): HResult;
begin
  Result := RegisterFromResource(FileName, MAKEINTRESOURCE(ID), sType, FALSE);
end;

function TRegObject.ResourceUnregisterSz(FileName: string; ID: string; sType: string): HResult;
begin
	if (ID = '') or (sType = '') then Exit(E_INVALIDARG);
	Result := RegisterFromResource(FileName, PWideChar(ID), sType, FALSE);
end;

function TRegObject.FileRegister(FileName: string): HResult;
begin
	Result := CommonFileRegister(FileName, TRUE);
end;

function TRegObject.FileUnregister(FileName: string): HResult;
begin
	Result := CommonFileRegister(FileName, FALSE);
end;

function TRegObject.StringRegister(Data: string): HResult;
begin
	Result := RegisterWithString(Data, TRUE);
end;

function TRegObject.StringUnregister(Data: string): HResult;
begin
	Result := RegisterWithString(Data, FALSE);
end;

function TRegObject.RegisterWithString(Data: string; Register: boolean): HResult;
var
  parser : TRegParser;
begin
	parser := TRegParser.Create(self);
  try
	  Result := parser.RegisterBuffer(Data, Register);
  finally
    parser.Free;
  end;
end;

function TRegObject.ClearReplacements(): HResult;
begin
	RepMap.Clear;
	Result := S_OK;
end;

function TRegObject.StrFromMap(Key: string): string;
begin
  Result := RepMap.Values[Key];
end;

function TRegObject.CommonFileRegister(FileName: string; Register: boolean): HResult;
var
  parser : TRegParser;
  temp : TStringList;
begin
	parser := TRegParser.Create(self);
  try
  	temp := TStringList.Create;
    try
      temp.LoadFromFile(FileName);
    	Result := parser.RegisterBuffer(temp.Text, Register);
    finally
      temp.Free;
    end;
  finally
    Parser.Free;
  end;
end;

function TRegObject.GenerateError(val : cardinal): HResult;
begin
  Result := MakeResult(SEVERITY_ERROR,FACILITY_WIN32,val);
end;

class function TRegParser.VTFromRegType(ValueType : string; var vt : TVarType) : boolean;
type
	typemap = record
		sz : string;
		vt : TVarType;
  end;
const
	map : array [0..3] of typemap = (
		(sz:StringVal; vt:VT_BSTR),
		(sz:multiStringVal; vt:VT_BSTR or VT_BYREF),
		(sz:DwordVal;  vt:VT_UI4),
		(sz:BinaryVal; vt:VT_UI1));
var
  i : integer;
begin
	for i := 0 to 3 do
  begin
		if (UpperCase(ValueType) = map[i].sz) then
		begin
			vt := map[i].vt;
			Exit(true);
		end;
	end;
	Result := FALSE;
end;

class function TRegParser.ChToByte(const ch : Char) : byte;
begin
	case (ch) of
		'0'..'9': Result := Ord(ch) - Ord('0');

		'A'..'F': Result := 10 +  Ord(ch) - Ord('A');
		'a'..'f': Result := 10 + Ord(ch) - Ord('a');
	else
  	Result := 0;
	end;
end;

class function TRegParser.HKeyFromString(Token : string) : HKey;
type
	keymap = record
		sz : string;
		key : HKEY;
	end;
const
	map : array [0..13] of keymap = (
		(sz:'HKCR'; key:HKEY_CLASSES_ROOT),
		(sz:'HKCU'; key:HKEY_CURRENT_USER),
		(sz:'HKLM'; key:HKEY_LOCAL_MACHINE),
		(sz:'HKU';  key:HKEY_USERS),
		(sz:'HKPD'; key:HKEY_PERFORMANCE_DATA),
		(sz:'HKDD'; key:HKEY_DYN_DATA),
		(sz:'HKCC'; key:HKEY_CURRENT_CONFIG),
		(sz:'HKEY_CLASSES_ROOT'; key:HKEY_CLASSES_ROOT),
		(sz:'HKEY_CURRENT_USER'; key:HKEY_CURRENT_USER),
		(sz:'HKEY_LOCAL_MACHINE'; key:HKEY_LOCAL_MACHINE),
		(sz:'HKEY_USERS'; key:HKEY_USERS),
		(sz:'HKEY_PERFORMANCE_DATA'; key:HKEY_PERFORMANCE_DATA),
		(sz:'HKEY_DYN_DATA'; key:HKEY_DYN_DATA),
		(sz:'HKEY_CURRENT_CONFIG'; key:HKEY_CURRENT_CONFIG));
var
  i : integer;
begin
	for i :=0 to 13 do
  begin
		if Token = map[i].sz then
			Exit(map[i].key);
	end;
	Result := 0;
end;

class function TRegParser.StrChr(sz : string; ch : Char) : PWideChar;
// Parses a string looking for ch and returns a pointer to it.
var
	p : PWideChar;
  s : PWideChar;
begin
  p := nil;
	if (sz = '') then Exit('');

  s := PWideChar(sz);
	while (s^ <> #0) do
	begin
		if (s^ = ch) then
    begin
			p := s;
			break;
		end;
		s := CharNext(s);
	end;
	Result := p;
end;

constructor TRegParser.Create(RegObj : TRegObject);
begin
  fRegObj := RegObj;
  Cur := nil;
end;

function TRegParser.IsSpace(ch : char) : boolean;
begin
	case ch of
		' ',#9,#13,#10: Exit(TRUE);
  end;
	Result := FALSE;
end;

procedure  TRegParser.SkipWhiteSpace;
begin
	while IsSpace(Cur^) do
		Cur := CharNext(Cur);
end;

function TRegParser.NextToken(var Token : string) : HResult;
var
  Prev : PWideChar;
  nChars : NativeInt;
  i : integer;
begin
  Token := '';
	SkipWhiteSpace;

	// NextToken cannot be called at EOS
	if (Cur^ = #0) then
		Exit(GenerateError(E_UNEXPECTED_EOS));

	// handle quoted value / key
	if (Quote = Cur^) then
	begin
		Cur := CharNext(Cur);

		while (Cur^ <> #0) and not EndOfVar do
		begin
			if (Quote = Cur^) then // If it is a quote that means we must skip it
				Cur := CharNext(Cur);

			Prev := Cur;
			Cur := CharNext(Cur);

			nChars := Cur - Prev;

			// Make sure we have room for nChars plus terminating NULL
      i := 0;
			while i < nChars do
      begin
				Token := Token + Prev^;
        Inc(i); Inc(Prev);
      end;
    end;

		if Cur^ = #0 then
		begin
			Exit(GenerateError(E_UNEXPECTED_EOS));
    end;

		Cur := CharNext(Cur);
  end else
	begin
		// Handle non-quoted ie parse up till first "White Space"
		while (Cur^ <> #0) and not IsSpace(Cur^) do
		begin
			Prev := Cur;
			Cur := CharNext(Cur);

			nChars := Cur - Prev;


      i := 0;
			while i < nChars do
      begin
				Token := Token + Prev^;
        Inc(i); Inc(Prev);
      end;
    end;
  end;
	Result := S_OK;
end;

function TRegParser.AddValue(rkParent : HKey; ValueName : string; var Token : string) : HResult;
var
	hr: HResult;
  szValue : string;
  szDestValue : string;
  vt : TVarType;
  lRes : integer;
  Len : integer;
  p,q,r : PWideChar;
  Val : cardinal;
  i, cbValue : cardinal;
  cbvalDiv2 : cardinal;
  rgbinary : PByte;
begin
	vt := VT_EMPTY;
	lRes := ERROR_SUCCESS;
  hr := NextToken(szValue);
	if FAILED(hr) then Exit(hr);

	if (not VTFromRegType(szValue, vt)) then
		Exit(GenerateError(E_TYPE_NOT_SUPPORTED));

	SkipWhiteSpace();
  hr := NextToken(szValue);
	if FAILED(hr) then Exit(hr);

	case vt of
  	VT_BSTR: begin
      lRes := RegSetValueEx(rkParent, PWideChar(ValueName), 0, REG_SZ, PChar(szValue),(Length(szValue) + 1) * SizeOf(Char));
    end;
  	VT_BSTR or VT_BYREF: begin
			Len := Length(szValue) + 2; //Allocate space for double null termination.
      SetLength(szDestValue,Len);
			p := PWideChar(szDestValue);
			q := PWideChar(szValue);
  		while (q^ <> #0) do
			begin
				r := CharNext(q);
				if (q^ = '\') and (r^ = #0) then
				begin
					Inc(p);
          p ^ := #0;
  				q := CharNext(r);
        end else
				begin
					p^ := q^;
  				Inc(p);
					Inc(q);
        end;
      end;
	    //Always terminate with 2 NULLs.
  		p^ := #0;
			Inc(p);
	    p^ := #0;
      lRes := RegSetValueEx(rkParent, PWideChar(ValueName), 0, REG_MULTI_SZ, PChar(szDestValue),(Length(szDestValue) + 1) * SizeOf(Char));
    end;
  	VT_UI4: begin
      Val := StrToIntDef(szValue,0);
      lRes := RegSetValueEx(rkParent, PWideChar(ValueName), 0, REG_DWORD, PChar(@Val), SizeOf(cardinal));
		end;
  	VT_UI1: begin
			cbValue := Length(szValue);
			if Odd(cbValue) then
				Exit(E_FAIL);  // data must be even number of bytes

			cbValDiv2 := cbValue div 2;
      rgBinary := AllocMem(cbValDiv2);
      try
  			for i := 0 to cbValue - 1 do
	  			rgBinary[(i div 2)] := rgBinary[i div 2] or (ChToByte(szValue[i])) shl (4*(1 - (i and 1)));
		  	lRes := RegSetValueEx(rkParent, PWideChar(ValueName), 0, REG_BINARY, rgBinary, cbValDiv2);
      finally
        FreeMem(rgbinary);
      end;
    end;
  end;

	if (lres <> ERROR_SUCCESS) then
	begin
		Exit(HResultFromWin32(lRes));
  end;

  hr := NextToken(Token);
	if FAILED(hr) then Exit(hr);

	Result := S_OK;
end;

function TRegParser.CanForceRemoveKey(Key : string) : boolean;
const
  cNeverDelete = 12;
  sNeverDelete : array [0..cNeverDelete -1] of string = ('AppID','CLSID',
             'Component Categories', 'FileType','Interface','Hardware','Mime',
             'SAM','SECURITY','SYSTEM','Software', 'TypeLib');
var
  NoDel : integer;
begin
	for NoDel := 0 to cNeverDelete - 1 do
		if Key = sNeverDelete[NoDel] then
			Exit(FALSE);                       // We cannot delete it

	Exit(TRUE);
end;

function TRegParser.HasSubKeys(key : HKEY) : boolean;
var
  cSubKeys : cardinal;
begin
	cSubKeys := 0;

	if (RegQueryInfoKey(key, nil, nil, nil,
							   @cSubKeys, nil, nil,
							   nil, nil, nil, nil, nil) <> ERROR_SUCCESS) then
		Exit(FALSE);

	Result := cSubKeys > 0;
end;

function TRegParser.HasValues(key : HKEY) : boolean;
const
  MAX_VALUE = 4096;
var
  cValues : cardinal;
  cMaxName : cardinal;
  ValueName : string;
begin
	cValues := 0;

	if RegQueryInfoKey(key, nil, nil, nil,
								  nil, nil, nil,
								  @cValues, nil, nil, nil, nil) <> ERROR_SUCCESS then
		Exit(FALSE);


	if (cValues = 1) then
	begin
		cMaxName := MAX_VALUE;
		SetLength(ValueName,MAX_VALUE);
		// Check to see if the Value is default or named
		if (RegEnumValue(key, 0, PWideChar(ValueName), cMaxName, nil, nil, nil, nil) = ERROR_SUCCESS)  and (ValueName <> '') then
			Exit(TRUE); // Named Value means we have a value
		Exit(FALSE);
	end;

	Result := cValues > 0; // More than 1 means we have a non-default value
end;

function TRegParser.SkipAssignment(var Token : string) : HResult;
var
	hr : HRESULT;
	szValue : string;
begin
	if (Token = cEquals) then
	begin
    hr := NextToken(Token);
		if FAILED(hr) then Exit(hr);

		// Skip assignment
		SkipWhiteSpace();
    hr := NextToken(szValue);
		if FAILED(hr) then Exit(hr);

    hr := NextToken(Token);
		if FAILED(hr) then Exit(hr);
  end;
	Result := S_OK;
end;

function TRegParser.PreProcessBuffer(Reg : string; var OutReg : string) : HResult;
var
  RedirectionEnabled: Boolean;
  NestingLevel : integer;
  RedirectionPresent: Boolean;
  InsideQuotes: Boolean;
  StartHKCU, EndHKCU : string;
  RootKey : PWideChar;
  Next : PWideChar;
  Len : integer;
  buf : string;
  sVar : string;
  ch : Char;
begin
	if (Reg = '') then Exit(E_POINTER);

	OutReg := '';
	Cur := PWideChar(Reg);
	Result := S_OK;

	RedirectionEnabled := fRegObj.PerUser;

	// nNestingLevel is used to avoid checking for unnecessary root key replacements
	// since all of them are expected to be at the top level.
	NestingLevel := 0;
	RedirectionPresent := false;
	InsideQuotes := false;

	while (Cur^ <> #0) do // look for end
	begin
		if RedirectionEnabled then
		begin
			StartHKCU := 'HKCU'#13#10'{'#9'Software'#13#10#9'{'#13#10#9#9'Classes';
			EndHKCU := #13#10#9'}'#13#10'}'#13#10;

			if NestingLevel = 0 then
      begin
				// Then we should be reading a root key. HKCR, HKCU, etc
				RootKey := TextPos(Cur, 'HKCR'); 	// if HKCR is found.
				if (RootKey <> nil) and	(RootKey = Cur) then	// if HKCR is the first token.
				begin
					// Skip HKCR
					Cur := CharNext(Cur);
					Cur := CharNext(Cur);
					Cur := CharNext(Cur);
					Cur := CharNext(Cur);

					// Add HKCU
					OutReg := OutReg + StartHKCU;
					RedirectionPresent := true;
				end;
			end;

			if ( Cur^ = Quote ) then
			begin
				if not InsideQuotes then
					InsideQuotes := true
				else
				begin
					// Make sure it is not an escaped sequence.
					if(EndOfVar) then
						InsideQuotes := false
					else
					begin
						// An escaped single quote...
						Cur := CharNext(Cur);
            ch := Cur^;
						OutReg := OutReg + ch;
					end;
        end;
      end;

			if not InsideQuotes and (Cur^ = '{') then
			begin
				Inc(NestingLevel);
			end;

			if not InsideQuotes and (Cur^ = '}') then
			begin
				Dec(NestingLevel);
				if (NestingLevel = 0) and RedirectionPresent then
				begin
					OutReg := OutReg + EndHKCU;
					RedirectionPresent := false;
				end;
			end;
		end;

		if (Cur^ = '%') then
		begin
			Cur := CharNext(Cur);
			if (Cur^ = '%') then
			begin
				ch := Cur^;
        OutReg := OutReg + ch;
			end else
			begin
				Next := StrScan(Cur, '%');
				if (Next = nil) then
				begin
					Result := GenerateError(E_UNEXPECTED_EOS);
					break;
				end;
				if ((Next- Cur) > 31) then
				begin
					Result := E_FAIL;
					break;
				end;
				Len := Next - Cur;
				buf := Copy(Cur,1,Len);
				sVar := fRegObj.StrFromMap(buf);
				if (sVar = '') then
				begin
					Result := GenerateError(E_NOT_IN_MAP);
					break;
				end;
				OutReg := OutReg + sVar;
				while (Cur <> Next) do
					Cur := CharNext(Cur);
			end;
		end else
		begin
      ch := Cur^;
			OutReg := OutReg + ch;
    end;
		Cur := CharNext(Cur);
	end;
end;

function TRegParser.RegisterBuffer(Buffer : string; Register : boolean) : HResult;
var
 	Token : string;
  sReg : string;
  hkBase : HKey;
  RegAtRegister : string;
begin
	sReg := '';
	Result := PreProcessBuffer(Buffer, sReg);
	if FAILED(Result) then Exit;

	Cur := PWideChar(sReg);

	// Preprocess szReg
	while (Cur^ <> #0) do
	begin
    Result := NextToken(Token);
		if FAILED(Result) then break;
    hkBase := HKeyFromString(Token);
		if (hkBase = 0) then
		begin
			Result := GenerateError(E_BAD_HKEY);
			break;
		end;

    Result := NextToken(Token);
		if FAILED(Result) then break;

		if (LeftBracket <> Token) then
		begin
			Result := GenerateError(E_MISSING_OPENKEY_TOKEN);
			break;
		end;
		if (Register) then
		begin
			RegAtRegister := Cur;
			Result := RegisterSubkeys(Token, hkBase, Register);
			if FAILED(Result) then
			begin
				Cur := PWideChar(RegAtRegister);
				RegisterSubkeys(Token, hkBase, false);
				break;
			end;
    end else
		begin
      Result := RegisterSubkeys(Token, hkBase, Register);
			if FAILED(Result) then break;
		end;

		SkipWhiteSpace;
  end;
end;

function RegDeleteKeyCompletely(Key: HKEY; SubKey: string): boolean;
var
  hSubKey: HKEY;
  error: Integer;
  cbName: cardinal;
  name: string;
  numSubKeys: cardinal;
  i: Integer;
begin
  error := RegOpenKeyEx(Key, PChar(SubKey), 0, KEY_QUERY_VALUE or
    KEY_ENUMERATE_SUB_KEYS or KEY_CREATE_SUB_KEY or KEY_SET_VALUE, hSubKey);
  if (error = ERROR_FILE_NOT_FOUND) then
  begin
    Result := true;
    exit;
  end;

  if (error <> ERROR_SUCCESS) then
  begin
    Result := False;
    exit;
  end;

  SetLength(name, 256);
  Result := true;
  if RegQueryInfoKey(hSubKey, nil, nil, nil, @numSubKeys, nil, nil, nil, nil,
    nil, nil, nil) = ERROR_SUCCESS then
  begin
    if (numSubKeys > 0) then
    begin
      for i := numSubKeys - 1 downto 0 do
      begin
        cbName := 256;
        error := RegEnumKeyEx(hSubKey, i, PChar(name), cbName, nil, nil,
          nil, nil);
        if error = ERROR_SUCCESS then
        begin
          if not RegDeleteKeyCompletely(hSubKey, name) then
          begin
            Result := False;
            break;
          end;
        end;
      end;
    end;
  end;
  RegCloseKey(hSubKey);
  if (Result) then
  begin
    error := RegDeleteKey(Key, PChar(SubKey));
    Result := (error = ERROR_SUCCESS);
  end;
end;

function TRegParser.RegisterSubkeys(var Token : string; hkParent : HKEY; Register : boolean; Recover : boolean = FALSE) : HResult;
var
	keyCur : HKey;
	lRes : integer;
	szKey : string;
	bDelete : boolean;
	InRecovery : boolean;
  TokenDelete : boolean;
  rkParent : HKey;
  ValueName : string;
  bHasSubKeys : boolean;
label
  EndCheck;
begin
	InRecovery := Recover;
  Result := NextToken(Token);
	if FAILED(Result) then Exit;

	while (Token <> RightBracket) do // Continue till we see a }
	begin
		bDelete := true;
		TokenDelete := UpperCase(Token) = UpperCase(Delete);

		if (UpperCase(Token) = UpperCase(ForceRemove)) or TokenDelete then
		begin
      Result := NextToken(Token);
     	if FAILED(Result) then break;

			if (Register) then
			begin
				if StrChr(Token, DirSep) <> nil then
					Exit(GenerateError(E_COMPOUND_KEY));

				if (CanForceRemoveKey(Token)) then
				begin
					RegDeleteKeyCompletely(hkParent,Token);
				end;
				if (TokenDelete) then
				begin
          Result := NextToken(Token);
        	if FAILED(Result) then break;
          Result := SkipAssignment(Token);
        	if FAILED(Result) then break;
					goto EndCheck;
        end;
      end;
    end;

		if UpperCase(Token) = UpperCase(NoRemove) then
		begin
			bDelete := false;    // set even for register
      Result := NextToken(Token);
    	if FAILED(Result) then break;
    end;

		if UpperCase(Token) = UpperCase(ValToken) then // need to add a value to hkParent
		begin
      Result := NextToken(ValueName);
    	if FAILED(Result) then break;

      Result := NextToken(Token);
    	if FAILED(Result) then break;

			if (Token <> cEquals) then
				Exit(GenerateError(E_EXPECTING_EQUAL));

			if (Register) then
			begin
				Result := AddValue(hkParent, ValueName, Token);
				if FAILED(Result) then Exit;
				goto EndCheck;
			end else
			begin
				if not Recover and bDelete then
				begin
					// We have to open the key for write to be able to delete.
					lRes := RegOpenKeyEx(hkParent, nil, 0, KEY_WRITE, rkParent);
					if (lRes = ERROR_SUCCESS) then
					begin
						lRes := RegDeleteValue(rkParent,PWideChar(ValueName));
						if (lRes <> ERROR_SUCCESS) and (lRes <> ERROR_FILE_NOT_FOUND) then
						begin
							// Key not present is not an error
							Result := HresultFromWin32(lRes);
							break;
						end;
          end	else
					begin
						Result := HresultFromWin32(lRes);
						break;
          end;
        end;
        Result := SkipAssignment(Token);
      	if FAILED(Result) then break;
				continue;  // can never have a subkey
			end;
    end;

		if (StrChr(Token, DirSep) <> nil) then
			Exit(GenerateError(E_COMPOUND_KEY));

		if (Register) then
		begin
			lRes := RegOpenKeyEx(hkParent, PWideChar(Token),0, KEY_READ or KEY_WRITE, keyCur);
			if (ERROR_SUCCESS <> lRes) then
			begin
				// Failed all access try read only
				lRes := RegOpenKeyEx(hkParent, PWideChar(Token), 0, KEY_READ,KeyCur);
				if (ERROR_SUCCESS <> lRes) then
        begin
					// Finally try creating it
					lRes := RegCreateKeyEx(hkParent, PWideChar(Token),0, nil, REG_OPTION_NON_VOLATILE, KEY_READ or KEY_WRITE,nil,KeyCur,nil);
					if (lRes <> ERROR_SUCCESS) then
						Exit(HresultFromWin32(lRes));
        end;
      end;

      Result := NextToken(Token);
    	if FAILED(Result) then break;

			if (Token = cEquals) then
			begin
        Result := AddValue(keyCur, '', Token);
				if FAILED(Result) then break; // NULL == default
      end;
    end else //Unregister
		begin
			if not Recover then
			begin
 				lRes := RegOpenKeyEx(hkParent, PWideChar(Token),0, KEY_READ, keyCur);
      end else
				lRes := ERROR_FILE_NOT_FOUND;


			// Open failed set recovery mode
			if (lRes <> ERROR_SUCCESS) then
				Recover := true;

			// Remember Subkey
			szKey := Token;

      Result := NextToken(Token);
    	if FAILED(Result) then break;
      Result := SkipAssignment(Token);
    	if FAILED(Result) then break;

			if (Token = LeftBracket) and (Length(Token) = 1) then
      begin
				Result := RegisterSubkeys(Token, keyCur, Register, Recover);
				// In recover mode ignore error
				if (FAILED(Result) and not Recover) then break;
				// Skip the }
        Result := NextToken(Token);
      	if FAILED(Result) then break;
      end;

			Recover := InRecovery;

			if (lRes = ERROR_FILE_NOT_FOUND) then
				// Key already not present so not an error.
				continue;

			if (lRes <> ERROR_SUCCESS) then
			begin
				// We are recovery mode continue on errors else break
				if (Recover) then continue
				else
				begin
					Result := HresultFromWin32(lRes);
					break;
        end;
      end;

			// If in recovery mode
			if (Recover and HasSubKeys(keyCur)) then
			begin
				// See if the KEY is in the NeverDelete list and if so, don't
				if (CanForceRemoveKey(szKey) and bDelete) then
				begin
					// Error not returned since we are in recovery mode. The error that caused recovery mode is returned
					RegDeleteKeyCompletely(KeyCur, szKey);
        end;
				continue;
      end;

			bHasSubKeys := HasSubKeys(keyCur);
			lRes := RegCloseKey(keyCur);
			if (lRes <> ERROR_SUCCESS) then
			  Exit(HResultFromWin32(lRes));

			if (bDelete and not bHasSubKeys) then
			begin
				lRes := RegDeleteKey(hkParent,PWideChar(szKey));
				if (lRes <> ERROR_SUCCESS) then
				begin
					Result := HresultFromWin32(lRes);
					break;
        end;
      end;
    end;

EndCheck:

		if (Register) then
		begin
			if (Token = LeftBracket) and (Length(Token) = 1) then
			begin
				Result := RegisterSubkeys(Token, keyCur, Register, FALSE);
				if FAILED(Result) then break;
        Result := NextToken(Token);
				if FAILED(Result) then break;
      end;
    end;
  end;
end;

function TRegParser.GenerateError(val : UINT) : HResult;
begin
  Result := MakeResult(SEVERITY_ERROR,FACILITY_WIN32,val);
end;

function TRegParser.EndOfVar : boolean;
begin
  Result := (Quote = Cur^) and (Quote <> CharNext(Cur)^);
end;

function UpdateRegistryFromResource(hInst : HMODULE; Res : PWideChar;
	Register : boolean; MapEntries : TStrings; PerUser : boolean) : HResult;
var
  hRes : HResult;
  p : IRegistrar;
  i : integer;
  szType, ModuleQuote, sModule : string;
begin
  p := TRegObject.Create(PerUser);

	if (MapEntries <> nil) then
  begin
    for i := 0 to MapEntries.Count - 1 do
    begin
			p.AddReplacement(MapEntries.Names[i], MapEntries.ValueFromIndex[i]);
		end;
  end;

	sModule := GetModuleName(hInst);

	// Buffer Size is Multiplied by 2 because we are calling ReplaceSingleQuote
	if (hInst = GetModuleHandle(nil)) then // register as EXE
	begin
		// If Registering as an EXE, then we quote the resultant path.
		// We don't do it for a DLL, because LoadLibrary fails if the path is
		// quoted
		ModuleQuote := '"' + sModule + '"';
		hRes := p.AddReplacement('Module', ModuleQuote);
	end else
	begin
		hRes := p.AddReplacement('Module', sModule);
	end;

	if FAILED(hRes) then Exit(hRes);

	hRes := p.AddReplacement('Module_Raw', sModule);
  if FAILED(hRes) then Exit(hRes);

	szType := 'REGISTRY';
	if (IS_INTRESOURCE(Res)) then
	begin
		if (Register) then
			hRes := p.ResourceRegister(sModule, LOWORD(NativeUInt(Res)), szType)
		else
			hRes := p.ResourceUnregister(sModule, LOWORD(NativeUInt(Res)), szType);
	end else
	begin
		if (Register) then
			hRes := p.ResourceRegisterSz(sModule, Res, szType)
		else
			hRes := p.ResourceUnregisterSz(sModule, Res, szType);
	end;
	Result := hRes;
end;

function UpdateRegistryFromFile(hInst : HMODULE; Filename : string;
	Register : boolean; MapEntries : TStrings; PerUser : boolean) : HResult;
var
  hRes : HResult;
  p : IRegistrar;
  i : integer;
  ModuleQuote, sModule : string;
begin
  p := TRegObject.Create(PerUser);

	if (MapEntries <> nil) then
  begin
    for i := 0 to MapEntries.Count - 1 do
    begin
			p.AddReplacement(MapEntries.Names[i], MapEntries.ValueFromIndex[i]);
		end;
  end;

	sModule := GetModuleName(hInst);

	// Buffer Size is Multiplied by 2 because we are calling ReplaceSingleQuote
	if (hInst = GetModuleHandle(nil)) then // register as EXE
	begin
		// If Registering as an EXE, then we quote the resultant path.
		// We don't do it for a DLL, because LoadLibrary fails if the path is
		// quoted
		ModuleQuote := '"' + sModule + '"';
		hRes := p.AddReplacement('Module', ModuleQuote);
	end else
	begin
		hRes := p.AddReplacement('Module', sModule);
	end;

	if FAILED(hRes) then Exit(hRes);

	hRes := p.AddReplacement('Module_Raw', sModule);
  if FAILED(hRes) then Exit(hRes);

  if (Register) then
		hRes := p.FileRegister(Filename)
	else
		hRes := p.FileUnregister(FileName);

	Result := hRes;
end;

end.
